from numpy import linspace, unique
from math import floor, ceil


def group_1D_to_2D(edges):
    result = []
    prev = edges[0]
    for i in range(1, len(edges)):
        result.append([prev, edges[i]])
        prev = edges[i]
    return result

def group_2D_to_1D(edges):
    result = [edges[0][0]]
    for e in edges:
        result.append(e[1])
    return result

def degroup_2D(x, edges):
    """
    Transform list of observations between edges.
    Each element in edges must contain start end end.
    The length of edges must be equal to the length of x
    """
    assert len(x) == len(edges)

    data = []
    for i, xx in enumerate(x):
        if xx == 1:
            data += [(edges[i][1] + edges[i][0]) / 2] * xx
        else:
            data += linspace(edges[i][0], edges[i][1], xx).tolist()
    return data


def degroup(x, groups):
    """
    Transform grouped data to ungrouped data assuming that data is evenly distributed in each interval.

    Parameters
    ==========

    x : list
        1D or 2D list of data to be ungrouped

    groups : list
        List of edge points in each interval. Each element can consist of a single or two edge values.
        Can either be a single list for all lists of observations or individual lists for each.

    Possible groups
    ==========
    One grouping, single edges. E.g. 3 intervals with distance 1:
        [1,2,3,4]

    One grouping, double edges. E.g. 3 intervals with distance 1:
        [[1,2], [2,3], [3,4]]

    Multiple groupings, single edges. E.g. 2 groupings with 3 intervals with distance 1:
        [[1,2,3,4], [1,2,3,4]]

    Multiple grouping, double edges. E.g. 2 groupings with 3 intervals with distance 1:
        [[[1,2], [2,3], [3,4]], [[1,2], [2,3], [3,4]]]
    """

    data = []
    if hasattr(groups[0], "__iter__"):
        if hasattr(groups[0][0], "__iter__"):
            assert len(groups) == len(x)
            for i, xx in enumerate(x):
                data.append(degroup_2D(xx, groups[i]))
            return data

        if len(groups[0]) == 2:
            if hasattr(x[0], "__iter__"):
                assert len(x[0]) == len(groups)
                for xx in x:
                    data.append(degroup_2D(xx, groups))
                return data
            else:
                return degroup_2D(x, groups)
        else:
            assert len(x) == len(groups)
            for i, xx in enumerate(x):
                data.append(degroup_2D(xx, group_1D_to_2D(groups[i])))
            return data

    if hasattr(x[0], "__iter__"):
        assert len(x[0]) + 1 == len(groups)
        for xx in x:
            data.append(degroup_2D(xx, group_1D_to_2D(groups)))
        return data
    else:
        return degroup_2D(x, group_1D_to_2D(groups))


def group(x: list, groups):
    """
    Group observations in intervals open on the lower limit, i.e. ]a,b] where a <= b
    """
    if not hasattr(groups[0], "__iter__"):
        groups = group_1D_to_2D(groups)

    x.sort()
    idx = 0
    outliers = 0
    while idx < len(x) and x[idx] <= groups[0][0]:
        outliers += 1
        idx += 1

    result = [0] * len(groups)
    for i, g in enumerate(groups):
        while idx < len(x) and x[idx] <= g[1]:
            result[i] += 1
            idx += 1

    outliers += len(x) - idx - 1
    if outliers > 0:
        Warning(f"{outliers} observations were no inside group intervals.")
    return result


class FrequencyTable:
    observation: list
    total: int
    hyppighed: list
    frekvens: list
    kumuleret_frekvens: list

    def __init__(self, x: list):
        self.total = len(x)
        hyp = unique(x, return_counts=True)
        self.observation = hyp[0].tolist()
        self.hyppighed = hyp[1].tolist()
        self.init_freq()

    def init_freq(self):
        self.frekvens = []
        self.kumuleret_frekvens = []
        kf = 0
        for i in range(len(self.observation)):
            f = self.hyppighed[i] / self.total * 100
            self.frekvens.append(f)
            kf += f
            self.kumuleret_frekvens.append(kf)

    def __str__(self):
        names = ["Observation", "Hyppighed", "Frekvens %", "Kumuleret frekvens %"]
        s = ""
        for n in names:
            s += f"{n} | "
        s += "\n"
        for i in range(len(self.observation)):
            for k, n in enumerate([self.observation, self.hyppighed, self.frekvens, self.kumuleret_frekvens]):
                if hasattr(n[i], "__iter__") and len(n[i]) == 2:
                    s += f"]{n[i][0]:>{floor((len(names[k])-3)/2)}.1f};{n[i][1]:>{ceil((len(names[k])-3)/2)}.1f}] | "
                elif type(n[i]) == float:
                    s += f"{n[i]:>{len(names[k])}.3f} | "
                else:
                    s += f"{n[i]:>{len(names[k])}} | "
            s += "\n"
        return s


class FrequencyTableGrouped(FrequencyTable):
    def __init__(self, x: list, groups: list):
        self.total = sum(x)
        self.observation = []
        self.hyppighed = x

        if hasattr(groups[0], "__iter__"):
            assert len(groups[0]) == 2
            assert len(groups) == len(x)
            for g in groups:
                self.observation.append(g)
        else:
            assert len(groups) == len(x) + 1
            for i in range(len(groups) - 1):
                self.observation.append([groups[i], groups[i + 1]])

        self.init_freq()


def frequency(x, groups=[], show=True):
    """
    Calculate count, frequency and cumulative frequency of grouped or ungrouped data
    """
    assert not hasattr(x[0], "__iter__")
    if groups == []:
        table = FrequencyTable(x)
    else:
        table = FrequencyTableGrouped(x, groups)
    if show:
        print(table)
    return table


if __name__ == "__main__":
    # fig, ax = boxplot([[8,3,2,1,1,2,3,4],[1,2,3,10]],labels=["test1","test2"])
    # x = degroup([[1, 2, 3, 8], [1, 2, 3, 10]], [0, 1, 2, 3, 10])
    # boxplot(x, labels=["test1", "test2"])
    t = frequency([8, 3, 2, 1, 1, 2, 3, 4])
    # print(t.frekvens)
    # print(t.hyppighed)
    # print(t.observation)
    # print(t.kumuleret_frekvens)
    t = frequency([8, 3, 2, 1], [1, 2, 3, 4, 5])
    t = frequency([8, 3, 2, 1], [[1, 2], [2, 3], [3, 4], [4, 5]])
    # print(t.observation)
